import os
import unittest

import numpy as np
import numpy.testing as npt
import pandas as pd
import xarray.testing as xrt

from tinamit.envolt.mds import ModeloVensimDLL, ModeloPySD, ModeloDS, gen_mds, \
    ErrorNoInstalado
# Los tipos de modelos DS que queremos comprobar.
from tinamit.tiempo.tiempo import EspecTiempo

tipos_modelos = {
    'PySD_Vensim': {'envlt': ModeloPySD, 'prueba': 'recursos/mds/prueba_senc_mdl.mdl'},
    'PySD_XMILE': {'envlt': ModeloPySD, 'prueba': 'recursos/mds/prueba_senc_xml.xmile'},
    'PySD_Py': {'envlt': ModeloPySD, 'prueba': 'recursos/mds/prueba_senc_py.py'},
    'dllVensim': {'envlt': ModeloVensimDLL, 'prueba': 'recursos/mds/prueba_senc_vpm.vpm'}
}

# Agregar la ubicación del fuente actual
dir_act = os.path.split(__file__)[0]
for d_m in tipos_modelos.values():
    d_m['prueba'] = os.path.join(dir_act, d_m['prueba'])


def generar_modelos_prueba():
    mods = {}
    for nmb, dic in tipos_modelos.items():
        cls = dic['envlt']
        if cls.instalado():
            mods[nmb] = cls(dic['prueba'])

    return mods


class TestLeerModelos(unittest.TestCase):
    """
    Verifica el funcionamiento de los programas de mds.
    """

    @classmethod
    def setUpClass(cls):

        # Generar las instancias de los modelos
        cls.modelos = generar_modelos_prueba()

        # Información sobre los variables del modelo de prueba
        cls.info_vars = {
            'Lluvia': {'unidades': 'm3/mes', 'líms': (0, np.inf)},
            'Nivel lago inicial': {'unidades': 'm3', 'líms': (0, np.inf)},
            'Flujo río': {'unidades': 'm3/mes', 'líms': (0, np.inf)},
            'Lago': {'unidades': 'm3', 'líms': (0, np.inf)},
            'Evaporación': {'unidades': 'm3/mes', 'líms': (0, np.inf)},
            'Aleatorio': {'unidades': 'Sdmn', 'líms': (0, 1)},
        }

    def test_leer_vars(símismo):
        """
        Comprobar que los nombres de los variables se leyeron correctamente.
        """
        for ll, mod in símismo.modelos.items():
            with símismo.subTest(mod=ll):
                vars_modelo = {str(vr) for vr in mod.variables}
                símismo.assertSetEqual(set(símismo.info_vars), vars_modelo)

    def test_unid_tiempo(símismo):
        """
        Comprobar que las unidades de tiempo se leyeron correctamente.
        """
        for ll, mod in símismo.modelos.items():
            with símismo.subTest(mod=ll):
                símismo.assertEqual('mes', mod.unidad_tiempo())

    def test_leer_info(símismo):
        """
        Comprobar que la documentación de cada variable se leyó correctamente.
        """

        for ll, mod in símismo.modelos.items():
            with símismo.subTest(mod=ll):
                símismo.assertTrue(len(mod.variables[v].info) > 0 for v in mod.variables)

    def test_leer_unidades(símismo):
        """
        Comprobar que las unidades de los variables se leyeron correctamente.
        """

        unids = {v: d_v['unidades'].lower() for v, d_v in símismo.info_vars.items()}

        for ll, mod in símismo.modelos.items():
            with símismo.subTest(mod=ll):
                unids_mod = {str(v): mod.variables[v].unid.lower() for v in mod.variables}
                símismo.assertDictEqual(unids, unids_mod)

    def test_leer_líms(símismo):
        """
        Comprobar que los límites de los variables se leyeron correctamente.
        """

        unids = {v: d_v['líms'] for v, d_v in símismo.info_vars.items()}

        for ll, mod in símismo.modelos.items():
            with símismo.subTest(mod=ll):
                unids_mod = {str(v): mod.variables[v].líms for v in mod.variables}
                símismo.assertDictEqual(unids, unids_mod)

    @classmethod
    def tearDownClass(cls):
        """
        Limpiar todos los archivos temporarios.
        """

        limpiar_mds()


class TestSimular(unittest.TestCase):
    @classmethod
    def setUpClass(cls):

        # Generar las instancias de los modelos
        cls.modelos = generar_modelos_prueba()
        cls.vals_inic = {
            'Nivel lago inicial': 1450,
            'Aleatorio': 2.3,
        }

        cls.res = {}
        for nmb, mod in cls.modelos.items():
            # Correr el modelo para 200 pasos, guardando los egresos del variable "Lago"
            cls.res[nmb] = mod.simular(
                t=200, extern=cls.vals_inic, vars_interés=['Lago', 'Aleatorio', 'Nivel lago inicial']
            )

    def test_cmb_vals_inic_constante_en_resultados(símismo):
        """
        Comprobar que los valores iniciales se establecieron correctamente en los resultados.
        """

        for nmb, res in símismo.res.items():
            v = 'Nivel lago inicial'

            with símismo.subTest(mod=nmb):
                npt.assert_array_equal(res[v].vals, símismo.vals_inic[v])

    def test_cambiar_vals_inic_var_dinámico(símismo):
        """
        Asegurarse que los valores iniciales de variables dinámicos aparezcan en el paso 0 de los resultados.
        """

        for nmb, res in símismo.res.items():
            v = 'Aleatorio'
            with símismo.subTest(mod=nmb):
                npt.assert_allclose(res[v].vals[0].values, símismo.vals_inic[v], rtol=0.0001)

    def test_cambiar_vals_inic_nivel(símismo):
        """
        Comprobar que valores iniciales pasados a un variable de valor inicial aparezcan en los resultados también.
        """

        for nmb, res in símismo.res.items():
            with símismo.subTest(mod=nmb):
                símismo.assertEqual(
                    res['Lago'].vals.values[0],
                    símismo.vals_inic['Nivel lago inicial']
                )

    def test_resultados_simul(símismo):
        """
        Assegurarse que la simulación dió los resultados esperados.
        """

        for nmb, res in símismo.res.items():
            with símismo.subTest(mod=nmb):
                # Leer el resultado del último día de simulación pára el variable "Lago"
                val_simulado = res['Lago'].vals.values[-1]

                # Debería ser aproximativamente igual a 100
                símismo.assertEqual(np.round(val_simulado, 3), 100)

    def test_simul_con_paso_2(símismo):

        for nmb, mod in símismo.modelos.items():
            with símismo.subTest(mod=nmb):
                res_paso_1 = símismo.res[nmb]['Lago'].vals.values[::2]
                res_paso_2 = mod.simular(
                    t=EspecTiempo(100, tmñ_paso=2), extern=símismo.vals_inic, vars_interés=['Lago']
                )['Lago'].vals.values
                npt.assert_allclose(res_paso_2, res_paso_1, rtol=0.001)

    def test_simul_guardar_cada_2(símismo):

        for nmb, mod in símismo.modelos.items():
            with símismo.subTest(mod=nmb):
                res_paso_1 = símismo.res[nmb]['Lago'].vals.values[::2]
                res_paso_2 = mod.simular(
                    t=EspecTiempo(200, guardar_cada=2), extern=símismo.vals_inic, vars_interés=['Lago']
                )['Lago'].vals.values
                npt.assert_equal(res_paso_1, res_paso_2)

    @classmethod
    def tearDownClass(cls):
        """
        Limpiar todos los archivos temporarios.
        """

        limpiar_mds()


class TestSimulExpres(unittest.TestCase):

    def test_simul_exprés(símismo):
        for nmb, mod in generar_modelos_prueba().items():
            with símismo.subTest(mod=nmb):
                extern = pd.DataFrame({'Lluvia': np.arange(10)}, index=np.arange(10))
                res_exprés = mod.simular(12, extern=extern)
                mod._correr_hasta_final = lambda: None
                res_por_paso = mod.simular(12, extern=extern)

                for res_var_exp, res_var_paso in zip(res_exprés, res_por_paso):
                    xrt.assert_equal(res_var_exp.vals, res_var_paso.vals)

    @classmethod
    def tearDownClass(cls):
        """
        Limpiar todos los archivos temporarios.
        """

        limpiar_mds()


class TestGenerarMDS(unittest.TestCase):
    """
    Verifica el funcionamiento del generado automático de modelos DS.
    """

    def test_generación_auto_mds(símismo):
        """
        Verificamos que funcione la generación automática de modelos DS a base de un fuente.
        """

        for m, d in tipos_modelos.items():
            with símismo.subTest(ext=os.path.splitext(d['prueba'])[1], envlt=d['envlt'].__name__):
                try:
                    mod = gen_mds(d['prueba'])  # Generar el modelo
                    símismo.assertIsInstance(mod, ModeloDS)
                except ErrorNoInstalado:
                    # No hay problema si el mds no se pudo leer en la computadora actual. De pronto no estaba instalado.
                    pass

    def test_error_extensión(símismo):
        """
        Comprobar que extensiones no reconocidas devuelvan un error.
        """

        with símismo.assertRaises(ErrorNoInstalado):
            gen_mds('recursos/mds/Modelo con extensión no reconocida.வணக்கம்')

    def test_modelo_erróneo(símismo):
        """
        Asegurarse que un fuente erróneo devuelva un error.
        """

        with símismo.assertRaises(FileNotFoundError):
            gen_mds('Yo no existo.mdl')

    @classmethod
    def tearDownClass(cls):
        limpiar_mds()


def limpiar_mds(direc='./recursos/mds'):
    """
    Limpiamos todos los documentos temporarios generados por los programas de modelos DS.
    """
    for c in os.walk(direc):
        for a in c[2]:
            ext = os.path.splitext(a)[1]
            try:
                if ext in ['.2mdl', '.vdf', '.csv'] or (ext == '.py' and any(
                        os.path.isfile(os.path.join(direc, os.path.splitext(a)[0] + e))
                        for e in ['.xmile', '.xml', '.mdl']
                )):
                    os.remove(os.path.join(direc, a))

            except (FileNotFoundError, PermissionError):
                pass



import numpy as np
import xarray as xr

from tinamit.config import _


class Variable(object):
    """La clase más general para variables de modelos en Tinamït."""

    def __init__(símismo, nombre, unid, ingr, egr, inic=0, líms=None, info=''):
        """

        Parameters
        ----------
        nombre: str
            El nombre del variable.
        unid: str or None
            Las unidades del variable.
        ingr: bool
            Si es un ingreso al modelo.
        egr: bool
            Si es un egreso del modelo.
        inic: int or float or np.ndarray
            El valor inicial del modelo.
        líms: tuple
            Los límites del variable.
        info: str
            Descripción detallada del variable.
        """
        if not (ingr or egr):
            raise ValueError(_('Si no es variable ingreso, debe ser egreso.'))
        símismo.nombre = nombre
        símismo.unid = unid
        símismo.ingr = ingr
        símismo.egr = egr
        símismo.inic = _a_np(inic)
        símismo.dims = símismo.inic.shape
        símismo.líms = _proc_líms(líms)
        símismo.info = info

        símismo._val = símismo.inic.astype(float)

    def poner_val(símismo, val):
        """
        Establece el valor del variable.

        Parameters
        ----------
        val: int or float or np.ndarray
            El nuevo valor.

        """

        if isinstance(val, np.ndarray):
            existen = np.invert(np.isnan(val))  # No cambiamos nuevos valores que faltan
            símismo._val[existen] = val[existen]
        elif not np.isnan(val):
            símismo._val[:] = val

    def obt_val(símismo):
        """
        Devuelve el valor del variable.
        """
        return símismo._val  # para disuadir modificaciones directas a `símismo._val`

    def reinic(símismo):
        """
        Reinicializa el variable a su valor pre-simulación.
        """
        símismo._val[:] = símismo.inic

    def __iadd__(símismo, otro):
        símismo.poner_val(símismo._val + otro)
        return símismo

    def __imul__(símismo, otro):
        símismo.poner_val(símismo._val * otro)

    def __imod__(símismo, otro):
        símismo.poner_val(símismo._val % otro)

    def __ifloordiv__(símismo, otro):
        símismo.poner_val(símismo._val // otro)

    def __ipow__(símismo, otro):
        símismo.poner_val(símismo._val ** otro)

    def __str__(símismo):
        return símismo.nombre


def _a_np(val):
    if isinstance(val, xr.DataArray):
        val = val.values
    if isinstance(val, np.ndarray):
        if val.shape:
            return val
        return np.array([val])
    elif isinstance(val, (int, float, np.number)):
        return np.array([val])
    else:
        return np.array(val)


def _proc_líms(líms):
    if líms is None:
        return -np.inf, np.inf
    else:
        return -np.inf if líms[0] is None else líms[0], np.inf if líms[1] is None else líms[1]



import math
import unittest

from tinamit.calibs.sintx.ec import Ecuación


class TestAnalizarEc(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ecs = {
            'neg': ('y=-x', 2, -2),
            'mín': ('y=mín(0, x)', 1, 0),
            'máx': ('y=máx(x, 0)', 1, 1),
            'abs': ('y=abs(x)', -1, 1),
            'exp': ('y=exp(x)', 0, 1),
            'ent': ('y=ent(x*1.0001)', 4, 4),
            'rcd': ('y=rcd(x^2)', 3, 3),
            'ln': ('y=ln(x)', 1, 0),
            'log': ('y=log(x)', 10, 1),
            'sin': ('y=sin(x)', 1, math.sin(1)),
            'cos': ('y=cos(x)', 1, math.cos(1)),
            'tan': ('y=tan(x)', 1, math.tan(1)),
            'sinh': ('y=sinh(x)', 5, math.sinh(5)),
            'cosh': ('y=cosh(x)', 5, math.cosh(5)),
            'tanh': ('y=tanh(x)', 5, math.tanh(5)),
            'asin': ('y=asin(x)', 1, math.asin(1)),
            'acos': ('y=acos(x)', 1, math.acos(1)),
            'atan': ('y=atan(x)', 1, math.atan(1)),
            'si_sino': ('y=si_sino(1, x, -x)', 1, 1),
            '>': ('y=x>1', 1, False),
            '<': ('y=x<3', 1, True),
            '>=': ('y=x>=0', 1, True),
            '<=': ('y=x<=0', 1, False),
            '==': ('y=x==0', 0, True),
            '!=': ('y=x!=1', 1, False),
            '+': ('y=x+3', 3, 6),
            '-': ('y=5.0-x', 1, 4),
            '*': ('y=2*x', 2.5, 5),
            '^': ('y=2^x', 3, 8),
            '/': ('y=x/3', 9, 3)
        }

    def _verificar_ec(símismo, ec, val_x, resp):
        ec_py = ec.a_python(paráms=[])
        símismo.assertEqual(ec_py([], {'x': val_x}), resp)

    def test_ec_a_tx(símismo):
        for nmbr, (ec, x, y) in símismo.ecs.items():
            with símismo.subTest(nmbr):
                obj_ec = Ecuación(ec)
                tx = str(obj_ec)
                obj_ec_2 = Ecuación(tx, nombre='y')

                símismo._verificar_ec(obj_ec, val_x=x, resp=y)
                símismo._verificar_ec(obj_ec_2, val_x=x, resp=y)

    def test_sacar_vars(símismo):
        ec = Ecuación('y=a*x + sin(b) + si_sino(x > d, a/(exp(-b*x)), máx(2, c))')
        símismo.assertSetEqual(ec.variables(), {'a', 'b', 'c', 'd', 'x'})

    def test_ec_de_MDS(símismo):
        ec = Ecuación('y=IF THEN ELSE (MAX (x, 0)>a, 5, 3)', dialecto='vensim')
        símismo.assertEqual(ec.a_python(paráms=['a'])([1], {'x': 2}), 5)

    def test_ec_con_nombre(símismo):
        ec = Ecuación('x*3', nombre='y')
        ec2 = Ecuación('y=x*3')
        símismo.assertEqual(str(ec), str(ec2))
        símismo.assertEqual(ec.nombre, ec2.nombre)

    def test_ec_sin_nombre(símismo):
        símismo.assertIsNone(Ecuación('x*3').nombre)

    def test_ec_otras_ecs(símismo):
        ec_1 = 'y = x*a'
        ec_2 = '2*b'
        ec_3 = '3'
        ec = Ecuación(ec_1, otras_ecs={'a': ec_2, 'b': ec_3})
        símismo.assertEqual(ec.a_python(paráms=[])([], {'x': 2}), 12)

    def test_ec_nombres_equiv(símismo):
        ec = Ecuación('y=a*x + b', nombres_equiv={'x': 'X'})
        símismo.assertEqual(ec.a_python(paráms=['a', 'b'])([4, 5], {'X': 2}), 13)

    def test_sacar_args(símismo):
        ec = Ecuación('y=1/f(a, x, 2*b)')
        for i in [None, 1, 2]:
            with símismo.subTest(i=i):
                símismo.assertListEqual(ec.sacar_args_func('f'), ['a', 'x', '2*b'])
                símismo.assertListEqual(ec.sacar_args_func('f', i=1), ['a'])
                símismo.assertListEqual(ec.sacar_args_func('f', i=2), ['a', 'x'])

    def test_obt_coef_var(símismo):
        ec = Ecuación('y=1/f(a, b*x, 2*c)')
        símismo.assertEqual(ec.coef_de('x'), ('b', False))

    def test_obt_coef_var_neg(símismo):
        ec = Ecuación('y=1/f(a, -b*x, 2*c)')
        símismo.assertEqual(ec.coef_de('x'), ('b', False))

    def test_obt_coef_var_div_coef(símismo):
        ec = Ecuación('y=a+x/b')
        símismo.assertEqual(ec.coef_de('x'), ('b', True))

    def test_obt_coef_var_dos_veces(símismo):
        ec = Ecuación('y=a*x + máx(2, a*x)')
        símismo.assertEqual(ec.coef_de('x'), ('a', False))

    def test_obt_coef_var_dos_veces_incompat(símismo):
        ec = Ecuación('y=a*x + b*x')
        símismo.assertIsNone(ec.coef_de('x'))

    def test_obt_coef_var_dupl_coef(símismo):
        ec = Ecuación('y=a*x + a')
        símismo.assertIsNone(ec.coef_de('x'))

    def test_obt_coef_var_dos_veces_falta_coef(símismo):
        ec = Ecuación('y=a*x + x')
        símismo.assertIsNone(ec.coef_de('x'))

    def test_obt_coef_var_dos_veces_incompat_div_mul(símismo):
        ec = Ecuación('y=a*x + a/x')
        símismo.assertIsNone(ec.coef_de('x'))

    def test_obt_coef_múltiple(símismo):
        ec = Ecuación('y=a*(x+b+a)')
        símismo.assertIsNone(ec.coef_de('x'))

    def test_obt_coef_var_div(símismo):
        ec = Ecuación('y=a+b/x')
        símismo.assertEqual(ec.coef_de('x'), ('b', True))

    def test_obt_coef_ec(símismo):
        ec = Ecuación('y=a*(x+b)')
        símismo.assertEqual(ec.coef_de('y'), ('a', True))  # Inverso porque a y "y" son de lados opuestos del "="

    def test_obt_coef_ec_div_coef(símismo):
        ec = Ecuación('y=(x+b)/a')
        símismo.assertEqual(ec.coef_de('y'), ('a', False))

    def test_obt_coef_ec_div(símismo):
        ec = Ecuación('y=a/(x+b)')
        símismo.assertEqual(ec.coef_de('y'), ('a', True))

    def test_obt_coef_ec_múltiple(símismo):
        ec = Ecuación('y=a*(x+b+a)')
        símismo.assertIsNone(ec.coef_de('y'))



import os

from setuptools import setup, find_packages

directorio = os.path.split(os.path.realpath(__file__))[0]


def leer(arch):
    with open(os.path.join(directorio, arch), encoding='UTF-8') as d:
        return d.read().strip()


# Leer la versión de Tinamït
versión = leer(os.path.join('tinamit', 'versión.txt'))

# Lo que sigue es código un poco complicado pero necesario para manejar las traducciones del código y de sus
# mensajes al usuario. En gran parte, su necesidad es culpa del paquete Babel que no reconoce lenguas Mayas, entre
# otras lenguas muy útiles.


# Por fin, el código habitual de instalación.
setup(
    name='tinamit',
    version=versión,
    packages=find_packages(),
    url='https://tinamit.readthedocs.io',
    download_url='https://github.com/julienmalard/Tinamit',

    license='GNU GPL 3',
    author='Julien Jean Malard',
    author_email='julien.malard@mail.mcgill.ca',
    description='Conexión de modelos socioeconómicos (dinámicas de los sistemas) con modelos biofísicos.',
    long_description=leer('README.rst'),

    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'pillow',
        'pandas',
        'python_dateutil',
        'regex',
        'lark-parser',
        'pint',
        'pyshp',
        'taqdir',
        'pysd',
        'lxml',
        'xarray',
        'chardet',
        'spotpy',
        'ennikkai'
    ],
    setup_requires=['Babel', 'polib'],
    zip_safe=False,

    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3 :: Only',
    ],
    include_package_data=True
)



from ..mod import ResultadosSimul


class ResultadosConectado(ResultadosSimul):
    def __init__(símismo, nombre, t, modelos, vars_interés):
        super().__init__(nombre, t, vars_interés=[])
        símismo.res_vars = {
            str(m): m.variables.gen_res(nombre, t, [v for v in vars_interés if v in m.variables]) for m in modelos
        }

    def a_dic(símismo):
        return {m: v.a_dic() for m, v in símismo.res_vars.items()}

    def __iter__(símismo):
        for m in símismo.res_vars.values():
            for v in m:
                yield v



import numpy as np

from tinamit.mod import VariablesMod
from tinamit.mod.var import Variable
from ._envolt import ModeloBF


class ModeloImpaciente(ModeloBF):
    """
    La clase pariente para modelos que deben correr por varios pasos al mismo tiempo, es decir,
    :class:`~tinamit.envolt.bf._indeter.Indeterminado` y :class:`~tinamit.envolt.bf._deter.Determinado`.
    """

    def __init__(símismo, tmñ_ciclo, variables, nombre='bf'):
        símismo.paso_en_ciclo = tmñ_ciclo - 1
        símismo.tmñ_ciclo = tmñ_ciclo

        super().__init__(variables, nombre)

    def iniciar_modelo(símismo, corrida):
        símismo.paso_en_ciclo = símismo.tmñ_ciclo - 1
        super().iniciar_modelo(corrida)

    def unidad_tiempo(símismo):
        raise NotImplementedError


class VariablesModImpaciente(VariablesMod):
    """
    Representa los variables de un modelo :class:`~tinamit.envolt.bf._impac.Impaciente`.
    """

    def vars_paso(símismo):
        """
        Devuelve los variables por paso.

        Returns
        -------
        list

        """
        return [v for v in símismo if isinstance(v, VarPaso)]

    def act_paso(símismo, paso):
        """
        Actualizar el paso de los variables en el ciclo.

        Parameters
        ----------
        paso: int
            El paso actual en el ciclo.

        """
        for v in símismo.vars_paso():
            v.act_paso(paso=paso)


class VarPaso(Variable):
    """
    Un variable de un modelo :class:`~tinamit.envolt.bf._impac.Impaciente` cuyo valor cambia con cada paso (y no
    solamente con cada ciclo).
    """

    def __init__(símismo, nombre, unid, ingr, egr, tmñ_ciclo, inic=0, líms=None, info=''):
        """

        Parameters
        ----------
        nombre: str
            El nombre del variable.
        unid: str or None
            Las unidades del variable.
        ingr: bool
            Si es un ingreso al modelo.
        egr: bool
            Si es un egreso del modelo.
        tmñ_ciclo: int
            El número de pasos en cada ciclo.
        inic: int or float or np.ndarray
            El valor inicial del modelo.
        líms: tuple
            Los límites del variable.
        info: str
            Descripción detallada del variable.
        """
        super().__init__(nombre, unid, ingr, egr, inic=inic, líms=líms, info=info)
        símismo._matr_paso = np.zeros((tmñ_ciclo, *símismo._val.shape))
        símismo._matr_paso[:] = símismo.inic
        símismo.paso = -1

    def poner_val(símismo, val):
        símismo._matr_paso[símismo.paso] = val
        super().poner_val(val)

    def obt_val(símismo):
        return símismo._matr_paso[símismo.paso]

    def poner_vals_paso(símismo, val, paso=None):
        """
        Establece el valor del variable a un paso dado. Si ``paso`` es ``None``, ``val`` debe ser una matriz donde
        eje 0 corresponde a todos los pasos del ciclo.

        Parameters
        ----------
        val: np.ndarray
            El nuevo valor.
        paso: int
            El paso al cual poner el nuevo valor del variable.

        """
        paso = paso if paso is not None else slice(None, None)

        # reformar para variables unidimensionales
        símismo._matr_paso[paso] = val.reshape(*símismo._matr_paso[paso].shape)

        # También hay que actualizar el variable paso
        super().poner_val(símismo._matr_paso[símismo.paso])

    def obt_vals_paso(símismo):
        """
        Obtener los valores del variable a todos los pasos del ciclo actual.

        Returns
        -------
        np.ndarray

        """
        return símismo._matr_paso

    def act_paso(símismo, paso):
        """
        Actualiza el paso actual del variable.

        Parameters
        ----------
        paso: int
            El nuevo paso en el ciclo actual.
        """
        símismo.paso = paso
        super().poner_val(símismo._matr_paso[paso])

    def reinic(símismo):
        símismo.paso = -1
        símismo._matr_paso[:] = símismo.inic
        super().reinic()



from ._impac import ModeloImpaciente, VarPaso, VariablesModImpaciente


class ModeloIndeterminado(ModeloImpaciente):
    """
    La clase pariente para todos modelos que avanzan por un número indeterminado de pasos a cada corrida.
    """

    def __init__(símismo, variables, nombre='bf'):
        super().__init__(tmñ_ciclo=1, variables=variables, nombre=nombre)

    def incrementar(símismo, rebanada):
        # Para simplificar el código un poco.
        p = símismo.paso_en_ciclo

        # Aplicar el incremento de paso
        p += rebanada.n_pasos

        # Si hay que avanzar el modelo externo, lanzar una su simulación aquí.
        while p >= símismo.tmñ_ciclo:
            p -= símismo.tmñ_ciclo

            # Avanzar la simulación
            símismo.tmñ_ciclo = símismo.mandar_modelo()

        # Guardar el pasito actual para la próxima vez.
        símismo.paso_en_ciclo = p

        # Actualizar el paso en los variables
        símismo.variables.act_paso(símismo.paso_en_ciclo)

        super().incrementar(rebanada)

    def unidad_tiempo(símismo):
        raise NotImplementedError

    def mandar_modelo(símismo):
        raise NotImplementedError


class VariablesModIndeterminado(VariablesModImpaciente):
    """
    Representa los variables de un modelo :class:`~tinamit.envolt.bf._indeter.ModeloIndeterminado`.
    """
    pass


class VarPasoIndeter(VarPaso):
    """
    Representa un variable de un modelo :class:`~tinamit.envolt.bf._indeter.ModeloIndeterminado` cuyo valor
    cambia a cada paso (y no solamente a cada ciclo).
    """

    def __init__(símismo, nombre, unid, ingr, egr, inic=0, líms=None, info=''):
        super().__init__(nombre, unid, ingr, egr, tmñ_ciclo=1, inic=inic, líms=líms, info=info)

    def poner_vals_paso(símismo, val, paso=None):
        # para hacer: ¿incecesario reimplementar esta función?        
        símismo._matr_paso = val



from tinamit.mod import VariablesMod
from tinamit.mod.var import Variable


class VariablesMDS(VariablesMod):
    """
    Representa los variables de un modelo :class:`~tinamit.envolt.mds._envolt.ModeloDS`.
    """
    def auxiliares(símismo):
        return [v for v in símismo if isinstance(v, VarAuxiliar)]

    def niveles(símismo):
        return [v for v in símismo if isinstance(v, VarNivel)]

    def constantes(símismo):
        return [v for v in símismo if isinstance(v, VarConstante)]

    def parientes(símismo, var):
        return símismo[str(var)].parientes

    def hijos(símismo, var):
        return [v for v in símismo if var in v.parientes]


class VarMDS(Variable):
    """
    Un variable de un modelo :class:`~tinamit.envolt.mds._envolt.ModeloDS`.
    """
    def __init__(símismo, nombre, unid, ingr, egr, ec, parientes, inic, subs=None, líms=None, info=''):
        super().__init__(nombre, unid, ingr=ingr, egr=egr, inic=inic, líms=líms, info=info)

        símismo.ec = ec
        símismo.parientes = parientes
        símismo.subs = subs


class VarConstante(VarMDS):
    """
    Un variable constante.
    """
    def __init__(símismo, nombre, unid, ec, parientes, inic, subs=None, líms=None, info=''):
        super().__init__(
            nombre, unid, ec=ec, parientes=parientes, inic=inic, subs=subs, líms=líms, info=info, ingr=True, egr=False
        )


class VarInic(VarMDS):
    """Un variable inicial."""
    def __init__(símismo, nombre, unid, ec, parientes, inic, subs=None, líms=None, info=''):
        super().__init__(
            nombre, unid, ec=ec, parientes=parientes, inic=inic, subs=subs, líms=líms, info=info, ingr=True, egr=False
        )


class VarNivel(VarMDS):
    """Un variable nivel."""
    def __init__(símismo, nombre, unid, ec, parientes, inic, subs=None, líms=None, info=''):
        super().__init__(
            nombre, unid, ec=ec, parientes=parientes, inic=inic, subs=subs, líms=líms, info=info, ingr=False, egr=True
        )
        símismo.var_inic = None

    def estab_var_inic(símismo, var_inic):
        símismo.var_inic = var_inic


class VarAuxiliar(VarMDS):
    """Un variable auxiliar."""
    def __init__(símismo, nombre, unid, ec, parientes, inic, subs=None, líms=None, info=''):
        super().__init__(
            nombre, unid, ec=ec, parientes=parientes, inic=inic, subs=subs, líms=líms, info=info, ingr=True, egr=True
        )



from ._auto import gen_mds, olvidar_envolt_mds, registrar_envolt_mds, ErrorNoInstalado
from ._envolt import ModeloDS
from ._vars import VariablesMDS, VarMDS, VarConstante, VarInic, VarNivel, VarAuxiliar
from .pysd import ModeloPySD
from .vensim_dll import ModeloVensimDLL



import json
import os
from itertools import chain as cadena

import babel
import pint
import pkg_resources
from pint import UnitRegistry
from tinamit.config import _
from tinamit.cositas import guardar_json, cargar_json

regu = UnitRegistry()
C_ = regu.Quantity

_archivo_trads = pkg_resources.resource_filename('tinamit.unids', 'trads_unids.json')
_archivo_pluriales = pkg_resources.resource_filename('tinamit.unids', 'pluriales.json')

l_dic_trads = None

if l_dic_trads is None:
    if os.path.isfile(_archivo_trads):
        try:
            l_dic_trads = cargar_json(_archivo_trads)
        except json.JSONDecodeError:  # pragma: sin cobertura
            l_dic_trads = []
    else:
        l_dic_trads = []

if os.path.isfile(_archivo_pluriales):
    _pluriales = cargar_json(_archivo_pluriales)
else:  # pragma: sin cobertura
    _pluriales = ['s', 'es', 'ें', 'கள்', 'க்கள்']
    guardar_json(_pluriales, _archivo_pluriales)


def act_arch_trads(l_d_t):
    """
    Actualiza el fuente de traducciones.

    Parameters
    ----------
    l_d_t : list[dict]

    """
    antes = hash(str(l_d_t))

    # UN conjunto vacío para las unidades presentes
    c_unids = set()

    # Agregar todas las unidades de Pint
    for d in dir(regu.sys):
        c_unids.update(u.lower() for u in dir(getattr(regu.sys, d)))

    # Todas las unidades ya en nuestro diccionario que podrían estar en Pint también.
    unids_doc = {d['en']['pr'].lower() for d in l_d_t if 'en' in d}

    # unids en Pint pero no en nuestro diccionario
    unids_faltan = c_unids.difference(unids_doc)

    # Lenguas ya incluidas en nuestro diccionario
    lengs = set(cadena.from_iterable([list(d) for d in l_d_t]))

    # Agregar las unidades de Pint que faltan
    for u in unids_faltan:  # pragma: sin cobertura
        d_u = {'en': {'pr': u, 'sn': []}}
        d_u.update({lng: {'pr': '', 'sn': []} for lng in lengs if lng != 'en'})
        l_d_t.append(d_u)

    # Limpiar el diccionario de traducciones
    for d in l_d_t:
        # Para cada unidad...
        for l, t in d.items():  # type: str, dict
            # Para cada lengua y diccionario de traducción...

            # Quitar nombres de unidades que estarían en la lista de sinónimos por error
            if t['pr'] in t['sn']:
                t['sn'].remove(t['pr'])

            d[l]['sn'] = list(set(t['sn']))  # Quitar sinónimos duplicados
            d[l]['sn'].sort()  # Ordenar los sinónimos

    # Guardar el diccionario de traducciones si hubieron modificaciones
    if hash(str(l_d_t)) != antes:
        try:
            guardar_json(obj=l_d_t, arch=_archivo_trads)
        except OSError:
            pass


def trad_unid(unid, leng_final, leng_orig=None, falla_silencio=True):
    """
    Traduce una unidad sencilla (no compuesta).

    Parameters
    ----------
    unid : str
        La unidad para traducir.
    leng_final : str
        La lengua a la cual traducir.
    leng_orig : str
        La lengua original de la unidad. Si no se especifica, se intentará adivinarla.
    falla_silencio: bool
        Si hay que devolver un error si no se encontró traducción.

    Returns
    -------
    str
        La unidad traducida.

    """

    # Quitar mayúsculas
    unid = unid.lower()

    l_u = buscar_singular(unid)

    # Buscar la unidad traducida
    unid_t = None
    if leng_orig is None:
        # Si no conocemos la lengua original, buscar todas.
        for u in l_u:
            unid_t = next((x[leng_final] for x in l_dic_trads
                           if (any([u == d['pr'] for d in x.values()]) or  # si `u` igual a la unidad principal
                               u in cadena.from_iterable([y['sn'] for y in x.values()])  # `u` igual a sinónimo
                               and leng_final in x  # ...y la lengua final existe
                               )
                           ), None)
            # Si encontramos la traducción, parar aquí.
            if unid_t is not None:
                break
    else:
        # Si conocemos la lengua original, solamente intentar las traducciones en esta lengua
        for u in l_u:
            unid_t = next(
                (x[leng_final] for x in l_dic_trads if  # El diccionario de la unidad traducida
                 (leng_orig in x  # Si la lengua original existe para esta unidad
                  and (u == x[leng_orig]['pr'] or u in x[leng_orig]['sn'])  # u igual a la unidad o sinónimos
                  and leng_final in x  # Y la lengua final también existe
                  )),
                None)

            # Si encontramos la traducción, parar aquí.
            if unid_t is not None:
                break

    # Si todavía no encontramos nada y la lengua original queda en inglés, ver si Pint nos pueda ayudar
    if unid_t is None and (leng_orig == 'en' or leng_orig is None):
        for u in l_u:
            try:
                u_base = regu.get_name(u)

                unid_t = next((x[leng_final] for x in l_dic_trads if
                               ('en' in x  # La traducción inglés existe
                                and (u_base in x['en']['sn'] or u_base == x['en']['pr'])  # La unidad corresponde
                                and leng_final in x)  # ...y la lengua final también existe
                               ), None)
            except pint.UndefinedUnitError:
                # Si no existía en Pint, seguir por el momento
                pass

    if unid_t is None or not len(unid_t):
        if falla_silencio:
            # Devolver la unidad no traducida si no encontramos nada
            unid_t = unid
        else:
            raise ValueError(_('No pudimos traducir "{u}" a la lengua "{leng}"').format(u=unid, leng=leng_final))
    else:
        # Devolver la traducción principal (no sinónimos) si encontramos algo
        unid_t = unid_t['pr']

    # Devolver la traducción
    return unid_t


def agregar_trad(unid, trad, leng_trad, leng_orig=None, guardar=True):
    """
    Agregar una traducción a una unidad.

    Parameters
    ----------
    unid : str
        La unidad original.
    trad : str
        La traducción de la unidad.
    leng_trad : str
        La lengua de la traducción.
    leng_orig : str
        La lengua original.
    guardar : bool
        Si hay que guardar la traducción para futuras sesiones de Python.

    """
    # Buscar el diccionario de la unidad
    d_unid = _buscar_d_unid(unid=unid, leng=leng_orig)

    # Si la lengua no existía, agragarla.
    if leng_trad not in d_unid:
        d_unid[leng_trad] = {'pr': '', 'sn': []}

    # Agregar la traducción
    if len(d_unid[leng_trad]['pr']):
        # Si ya existía traducción, agregar ésta como sinónima
        if trad not in d_unid[leng_trad]['sn']:
            d_unid[leng_trad]['sn'].append(trad)
    else:
        # Si no existía traducción, ésta será la versión principal de la unidad.
        d_unid[leng_trad]['pr'] = trad

    # Guardar si necesario.
    if guardar:
        act_arch_trads(l_dic_trads)


def agregar_sinónimos(unid, sinónimos, leng, guardar=False):
    """
    Agrega sinónimos a una unidad.

    Parameters
    ----------
    unid : str
        La unidad original.
    sinónimos : str | list
        Los sinónimos.
    leng : str
        La lengua.
    guardar : bool
        Si guardamos los sinónimos para futuras sesiones de Python.

    """
    # Buscar el diccionario de la unidad.
    d_unid = _buscar_d_unid(unid=unid, leng=leng)

    # Agregar los sinónimos
    if isinstance(sinónimos, str):
        sinónimos = [sinónimos]
    for s in sinónimos:
        if s not in d_unid[leng]['sn']:
            d_unid[leng]['sn'].append(s)

    # Guardar si necesario
    if guardar:
        act_arch_trads(l_dic_trads)


def _buscar_d_unid(unid, leng=None):
    """
    Busca el diccionario de una unidad.

    Parameters
    ----------
    unid : str
        El nombre de la unidad.
    leng : str
        La lengua de la unidad.

    Returns
    -------
    dict
        El diccionario de las traducciones de la unidad.

    """

    # Quitar mayúculas
    unid = unid.lower()

    l_u = buscar_singular(unid)
    d_unid = None

    for u in l_u:

        if leng is None:
            # Si no conocemos la lengua, buscar en todas las posibilidades.
            d_unid = next((x for x in l_dic_trads
                           if any(u in y['sn'] or u == y['pr'] for y in x.values())
                           ), None)
        else:
            # Si conocemos la lengua, buscar únicamente en las unidades de ésta.
            d_unid = next((x for x in l_dic_trads
                           if leng in x and (u in x[leng]['sn'] or u == x[leng]['pr'])
                           ), None)

        if d_unid is not None:
            # Si encontramos algo, para aquí.
            break

    # Si todavía no encontramos nada y la lengua original queda en inglés, ver si Pint nos pueda ayudar
    if d_unid is None and (leng == 'en' or leng is None):
        for u in l_u:
            try:
                u_base = regu.get_name(u)

                d_unid = next((x for x in l_dic_trads
                               if 'en' in x and (u_base in x['en']['sn'] or u_base == x['en']['pr'])
                               ), None)

            except pint.UndefinedUnitError:
                # Si no existía en Pint, seguir por el momento
                pass

    # Si todavía no encontramos nada, tenemos un error
    if d_unid is None:
        if leng is None:
            raise ValueError(_('La unidad "{}" no existe en cualquier lengua conocida.').format(unid))
        else:
            raise ValueError(_('La unidad "{u}" no existe en la lengua "{lng}".').format(u=unid, lng=leng))

    # Devolver el diccionario de unidad encontrado.
    return d_unid


def buscar_singular(u):
    """
    Busca las formas singulares posibles de una unidad.

    Parameters
    ----------
    u : str
        La unidad potencialmente plurial.

    Returns
    -------
    list[str]
        Una lista de las formas singulares potenciales.

    """

    # Generar una lista de la unidad y de potenciales formas singulares
    l_u = [u]
    for p in _pluriales:
        try:
            if u.endswith(p):
                l_u.append(u[:-len(p)])
        except IndexError:
            pass  # En el caso que la extensión plural sea más grande que la unidad sí misma.

    return l_u


# Actualizar las traducciones al importar este módulo
if os.path.getmtime(babel.__file__) > os.path.getmtime(_archivo_trads):
    act_arch_trads(l_d_t=l_dic_trads)
básicas = {
    'month': 'mes',
    'year': 'año',
    'day': 'día'
}
for ll, v in básicas.items():
    agregar_trad(ll, v, leng_trad='es', leng_orig='en')



import os

import numpy as np
import shapefile as sf
from matplotlib import colors, cm
from matplotlib.backends.backend_agg import FigureCanvasAgg as TelaFigura
from matplotlib.figure import Figure as Figura
from matplotlib.axes import Axes
import matplotlib.pyplot as dib
from tinamit.config import _
from tinamit.cositas import detectar_codif

from ..mod import ResultadosSimul, ResultadosGrupo


def dibujar_mapa(formas, archivo=None, título=None, fig=None):
    """
    Dibuja un mapa.

    Parameters
    ----------
    formas: list of Forma
        Las formas para incluir.
    archivo: str
        Dónde hay que guardar el gráfico. Si es ``None``, no se guardará el gráfico.
    título: str
        El título del mapa.
    fig: matplotlib.Figure
        Figura para dibujar el mapa.

    Returns
    -------
    tuple[Figure, Axes]
        La figura y sus ejes.
    """

    formas = [formas] if isinstance(formas, Forma) else formas

    if not fig:
        fig = Figura()
        TelaFigura(fig)
    ejes = fig.add_subplot(111)
    ejes.set_aspect('equal')

    for frm in formas:
        frm.dibujar(ejes, fig)
    if título is not None:
        ejes.set_title(título)
    if archivo:
        fig.savefig(archivo, dpi=500)

    return fig, ejes


def dibujar_mapa_de_res(
        forma_dinámica, res, var, t, escala=None, título='', directorio=None, otras_formas=None
):
    """
    Dibujar los resultados de una simulación en un mapa.

    Parameters
    ----------
    forma_dinámica: FormaDinámica
        La forma cuyos colores variarán según los resultados.
    res: ResultadosSimul or ResultadosGrupo
        Los resultados para dibujar.
    var: str
        El variable de interés.
    t: int or tuple or range or list
        Los tiempos a los cuales queremos graficar los resultados.
    escala: tuple
        El rango para aplicar colores. Si es ``None``, se aplicará según los datos de los resultados.
    título: str
        El título del gráfico.
    directorio: str
        Dónnde hay que guardar el gráfico.
    otras_formas: list of FormaEstática or FormaEstática
        Las otras formas (estáticas) para incluir en el gráfico.

    """
    título = título or res.nombre
    otras_formas = otras_formas or []
    if isinstance(otras_formas, FormaEstática):
        otras_formas = [otras_formas]

    def _extr_i(matr, i):
        return matr[i] if isinstance(i, int) else matr.sel(**{_('fecha'): i}).values

    if isinstance(res, ResultadosSimul):
        res_var = res[var].vals
        unids = res[var].var.unid

        escala = escala or (np.min(res_var.values), np.max(res_var.values))

        def _extr_res_i(i):
            return _extr_i(res_var, i)

    else:

        unids = list(res.values())[0][var].var.unid

        todos_vals = np.array([res_lg[var].vals for res_lg in res.values()])
        escala = escala or (np.min(todos_vals), np.max(todos_vals))

        def _extr_res_i(i):
            return {lg: _extr_i(res_lg[var].vals, i) for lg, res_lg in res.items()}

    if isinstance(t, int):
        pasos = range(t, t + 1)
    else:
        pasos = t

    for i_p in pasos:
        título_i = '{}_{}'.format(título, i_p)
        res_i = _extr_res_i(i_p)
        forma_dinámica.estab_valores(res_i, escala_valores=escala, unidades=unids)

        arch_i = os.path.join(directorio, título_i) if directorio is not None else None
        fig_i = None if directorio else dib.figure()
        dibujar_mapa([forma_dinámica, *otras_formas], archivo=arch_i, título=título_i, fig=fig_i)


class Forma(object):
    """
    Clase pariente para todas las formas que se pueden dibujar.
    """
    def __init__(símismo, archivo, llenar, alpha):
        codif = detectar_codif(os.path.splitext(archivo)[0] + '.dbf')
        símismo.forma = sf.Reader(archivo, encoding=codif)
        símismo.llenar = llenar
        símismo.alpha = alpha

    def dibujar(símismo, ejes, fig):
        """
        Agrega la forma a la figura.

        Parameters
        ----------
        ejes:
            Los ejes de la figura.
        fig:
            La figura.

        """
        raise NotImplementedError

    def _dibujar_frm(símismo, ejes, color):
        for i, frm in enumerate(símismo.forma.shapes()):
            puntos = frm.points
            partes = frm.parts

            for ip, i0 in enumerate(partes):  # Para cada parte del imagen

                if ip < len(partes) - 1:
                    i1 = partes[ip + 1] - 1
                else:
                    i1 = len(puntos)

                seg = puntos[i0:i1 + 1]
                x_lon = np.zeros((len(seg), 1))
                y_lat = np.zeros((len(seg), 1))
                for j in range(len(seg)):
                    x_lon[j] = seg[j][0]
                    y_lat[j] = seg[j][1]

                clr = color[i] if isinstance(color, np.ndarray) else color
                if símismo.llenar:
                    ejes.fill(x_lon, y_lat, color=clr, alpha=símismo.alpha)
                else:
                    ejes.plot(x_lon, y_lat, color=clr, alpha=símismo.alpha)


class FormaEstática(Forma):
    """
    Clase de base para formas estáticas en el mapa, cuyos colores no cambian.
    """

    def __init__(símismo, archivo, color, llenar, alpha):
        símismo.color = color
        super().__init__(archivo, llenar=llenar, alpha=alpha)

    def dibujar(símismo, ejes, fig):
        símismo._dibujar_frm(ejes, color=símismo.color)


class FormaDinámica(Forma):
    """
    Forma cuyos colores se asignan según valores numéricos.
    """

    def __init__(símismo, archivo, escala_colores=None, llenar=True, alpha=1):
        """

        Parameters
        ----------
        archivo: str
            El archivo ``.shp``.
        escala_colores: list or tuple or str or int or None
            Lista de dos colores para establecer una escala de colores. Si es un solo color, se agregará el color
            blanco. Si es ``-1``, se inverserán los colores automáticos.
        llenar: bool
            Si hay que llenar la forma o simplement delinear su contorno.
        alpha: float or int
            La opacidad del interior de la forma. Solamente aplica si ``llenar`` est ``False``.
        """
        super().__init__(archivo, llenar=llenar, alpha=alpha)

        símismo.escala_colores = símismo._resolver_colores(escala_colores)
        símismo.valores = np.full(len(símismo.forma.shapes()), np.nan)
        símismo.unidades = None
        símismo.escala = None

    def estab_valores(símismo, valores, escala_valores=None, unidades=None):
        """
        Establece los valores para colorar.

        Parameters
        ----------
        valores: np.ndarray or dict
            Los valores para dibujar. Debe ser del mismo tamaño que el archivo ``.shp`` en ``archivo``.
        escala_valores: tuple or list or None
            La escala para el rango de colores. Si es ``None``, se ajustará el rango según de los valores dados.
        unidades: str, optional
            Las unidades.
        """
        símismo.unidades = unidades

        símismo._llenar_valores(valores)

        if escala_valores is None:
            if np.all(np.isnan(símismo.valores)):
                escala_valores = (0, 1)
            else:
                escala_valores = (np.nanmin(símismo.valores), np.nanmax(símismo.valores))
                if escala_valores[0] == escala_valores[1]:
                    escala_valores = (escala_valores[0] - 0.5, escala_valores[0] + 0.5)

        símismo.escala = escala_valores

    def dibujar(símismo, ejes, fig):
        vals_norm = (símismo.valores - símismo.escala[0]) / (símismo.escala[1] - símismo.escala[0])

        d_clrs = _gen_d_mapacolores(colores=símismo.escala_colores)

        mapa_color = colors.LinearSegmentedColormap('mapa_color', d_clrs)
        norm = colors.Normalize(vmin=símismo.escala[0], vmax=símismo.escala[1])
        cpick = cm.ScalarMappable(norm=norm, cmap=mapa_color)
        cpick.set_array(np.array([]))

        v_cols = mapa_color(vals_norm)
        v_cols[np.isnan(vals_norm)] = 1

        símismo._dibujar_frm(ejes=ejes, color=v_cols)

        if símismo.unidades is not None:
            fig.colorbar(cpick, label=símismo.unidades)
        else:
            fig.colorbar(cpick)

    @staticmethod
    def _resolver_colores(colores):
        if colores is None:
            return ['#FF6666', '#FFCC66', '#00CC66']
        elif colores == -1:
            return ['#00CC66', '#FFCC66', '#FF6666']
        elif isinstance(colores, str):
            return ['#FFFFFF', colores]
        return colores

    def _extraer_col(símismo, col):
        nombres_attr = [field[0] for field in símismo.forma.fields[1:]]
        try:
            return [x.record[nombres_attr.index(col)] for x in símismo.forma.shapeRecords()]
        except ValueError:
            raise ValueError(_('La columna "{}" no existe en la base de datos.').format(col))

    def _llenar_valores(símismo, valores):
        raise NotImplementedError


class FormaDinámicaNumérica(FormaDinámica):
    """
    Forma dinámica cuyos valores se asignan a los polígonos de la forma ``.shp`` por su orden en la matriz de valores.
    """

    def __init__(símismo, archivo, col_id=None, escala_colores=None, llenar=True, alpha=1):
        """

        Parameters
        ----------
        archivo: str
            El archivo ``.shp``.
        col_id: str, optional
            La columna con el número de cada polígono en la forma ``.shp``. Si es ``None``, se asiñará número según
            su orden en la forma ``.shp``.
        escala_colores: list or tuple or str or int or None
            Lista de dos colores para establecer una escala de colores. Si es un solo color, se agregará el color
            blanco. Si es ``-1``, se inverserán los colores automáticos.
        llenar: bool
            Si hay que llenar la forma o simplement delinear su contorno.
        alpha: float or int
            La opacidad del interior de la forma. Solamente aplica si ``llenar`` est ``False``.
        """
        super().__init__(archivo, escala_colores, llenar, alpha)

        if col_id is not None:
            ids = np.array(símismo._extraer_col(col_id))
            símismo.ids = ids - np.min(ids)
        else:
            símismo.ids = None

    def _llenar_valores(símismo, valores):
        if símismo.ids is None:
            símismo.valores[:] = valores
        else:
            símismo.valores[:] = valores[símismo.ids]


class FormaDinámicaNombrada(FormaDinámica):
    """
    Forma dinámica cuyos valores se asignan a los polígonos de la forma ``.shp`` por su llave en el diccionario de
    valores.
    """

    def __init__(símismo, archivo, col_id, escala_colores=None, llenar=True, alpha=1):
        """

        Parameters
        ----------
        archivo: str
            La archivo ``.shp``.
        col_id: str
            La columna en el archivo ``.shp`` con el nomrbe de cada polígono.
        escala_colores: list or tuple or str or int or None
            Lista de dos colores para establecer una escala de colores. Si es un solo color, se agregará el color
            blanco. Si es ``-1``, se inverserán los colores automáticos.
        llenar: bool
            Si hay que llenar la forma o simplement delinear su contorno.
        alpha: float or int
            La opacidad del interior de la forma. Solamente aplica si ``llenar`` est ``False``.
        """
        super().__init__(archivo, escala_colores, llenar, alpha)
        símismo.ids = [str(x) for x in símismo._extraer_col(col_id)]

    def _llenar_valores(símismo, valores):
        símismo.valores[:] = np.nan
        for id_, val in valores.items():
            i = símismo.ids.index(id_)
            símismo.valores[i] = val


class Agua(FormaEstática):
    """
    Representa áreas de agua.
    """

    def __init__(símismo, archivo, llenar=True):
        """

        Parameters
        ----------
        archivo: str
            El archivo ``.shp``.
        llenar: bool
            Si hay que llenar el cuerpo de agua o no.
        """
        super().__init__(archivo=archivo, color='#13A6DD', llenar=llenar, alpha=0.5)


class Bosque(FormaEstática):
    """
    Representa áreas con bosque.
    """

    def __init__(símismo, archivo):
        super().__init__(archivo=archivo, color='#33A02C', llenar=True, alpha=0.7)


class Calle(FormaEstática):
    """
    Representa calles.
    """

    def __init__(símismo, archivo):
        super().__init__(archivo=archivo, color='#585763', llenar=False, alpha=1)


class Ciudad(FormaEstática):
    """
    Representa áreas urbanas.
    """

    def __init__(símismo, archivo):
        super().__init__(archivo=archivo, color='#FB9A99', llenar=True, alpha=1)


class OtraForma(FormaEstática):
    """
    Representa otras áreas no representadas por las otras formas disonibles.
    """

    def __init__(símismo, archivo):
        super().__init__(archivo=archivo, color='#FFECB3', llenar=True, alpha=1)


def _hex_a_rva(hx):
    """
    Convierte colores RVA a Hex.

    Parameters
    ----------
    hx: str
        El valor hex.

    Returns
    -------
    tuple
        El valor rva.
    """
    return tuple(int(hx.lstrip('#')[i:i + 2], 16) for i in (0, 2, 4))


def _gen_d_mapacolores(colores):
    """
    Genera un diccionario de mapa de color para MatPlotLib.

    Parameters
    ----------
    colores: list
        Una lista de colores

    Returns
    -------
    dict
        Un diccionario para MatPlotLib
    """

    clrs_rva = [_hex_a_rva(x) for x in colores]
    # noinspection PyTypeChecker
    n_colores = len(colores)

    dic_c = {'red': tuple((round(i / (n_colores - 1), 2), clrs_rva[i][0] / 255, clrs_rva[i][0] / 255) for i in
                          range(0, n_colores)),
             'green': tuple(
                 (round(i / (n_colores - 1), 2), clrs_rva[i][1] / 255, clrs_rva[i][1] / 255) for i in
                 range(0, n_colores)),
             'blue': tuple(
                 (round(i / (n_colores - 1), 2), clrs_rva[i][2] / 255, clrs_rva[i][2] / 255) for i in
                 range(0, n_colores))}

    return dic_c



import gettext
import json
import os
from warnings import warn as avisar

from pkg_resources import resource_filename

from tinamit.cositas import cargar_json, guardar_json


# Código para manejar configuraciones de Tinamït
class OpsConfig(object):
    def __init__(símismo, pariente, valores=None):
        valores = valores or {}
        símismo.valores = {ll: OpsConfig(símismo, v) if isinstance(v, dict) else v for ll, v in valores.items()}
        símismo.pariente = pariente

    def a_dic(símismo):
        return {ll: v.a_dic() if isinstance(v, OpsConfig) else v for ll, v in símismo.valores.items()}

    def borrar(símismo, llave):
        if isinstance(llave, str):
            símismo.valores.pop(llave)
        elif len(llave) == 1:
            símismo.valores.pop(llave[0])
        else:
            símismo.valores[llave[0]].borrar(llave[1:])

    def guardar(símismo):
        símismo.pariente.guardar()

    def __contains__(símismo, itema):
        return itema in símismo.valores

    def __getitem__(símismo, itema):
        if isinstance(itema, str):
            return símismo.valores[itema]
        elif len(itema) == 1:
            return símismo.valores[itema[0]]
        else:
            return símismo.valores[itema[0]][itema[1:]]

    def __setitem__(símismo, llave, valor):
        if isinstance(llave, str):
            símismo.valores[llave] = valor
        elif len(llave) == 1:
            símismo.valores[llave[0]] = valor
        else:
            if not llave[0] in símismo:
                símismo.valores[llave[0]] = OpsConfig(pariente=símismo)
            símismo.valores[llave[0]][llave[1:]] = valor
        símismo.guardar()


class Config(OpsConfig):

    def __init__(símismo, archivo, auto=None):
        auto = auto or {}

        símismo.archivo = archivo
        símismo.val_auto = auto.copy()

        try:
            val_arch = cargar_json(archivo)
        except (FileNotFoundError, json.decoder.JSONDecodeError, PermissionError):
            val_arch = {}

        auto.update(val_arch)

        super().__init__(pariente=None,valores=auto)
        try:
            símismo.guardar()
        except OSError:
            pass

    def reinic(símismo):
        símismo.valores = OpsConfig(pariente=símismo, valores=símismo.val_auto)
        símismo.guardar()

    def borrar(símismo, llave):
        super().borrar(llave)
        símismo.guardar()

    def guardar(símismo):
        try:
            guardar_json(símismo.a_dic(), símismo.archivo)
        except PermissionError:
            avisar(_('No se pudo guardar la configuración.'))

    def __setitem__(símismo, llave, valor):
        super().__setitem__(llave, valor)
        símismo.guardar()


class Trads(object):
    AUX = 'aux'
    PRINC = 'principal'
    DOMINIO = 'leng'

    def __init__(símismo, dir_local):
        símismo.dir_local = dir_local
        símismo._conf = conf[Trads.DOMINIO]

        símismo._trads = símismo._regen_trads()

    def idioma(símismo):
        return símismo._conf[Trads.PRINC]

    def auxiliares(símismo):
        return símismo._conf[Trads.AUX]

    def cambiar_idioma(símismo, idioma):
        símismo._conf[Trads.PRINC] = idioma
        if idioma in símismo.auxiliares():
            símismo.auxiliares().pop(idioma)  # aprovechamos del enlace dinámico

        símismo._trads = símismo._regen_trads()

    def cambiar_aux(símismo, aux):
        if isinstance(aux, str):
            aux = [aux]

        principal = símismo.idioma()
        aux = [i for i in aux if i != principal]
        símismo._conf[Trads.AUX] = aux

        símismo._trads = símismo._regen_trads()

    def trad(símismo, texto):
        return símismo._trads.gettext(texto)

    def _regen_trads(símismo):
        return gettext.translation(
            __name__, localedir=símismo.dir_local, fallback=True,
            languages=[símismo.idioma()] + símismo.auxiliares()
        )


conf = Config(
    resource_filename('tinamit', 'config.json'),
    auto={
        Trads.DOMINIO: {Trads.PRINC: 'cak', Trads.AUX: ['es']},
        'envolt': {}
    }
)

conf_mods = conf['envolt']

trads = Trads(
    dir_local=os.path.join(os.path.abspath(os.path.dirname(__file__)), '_local', '_fuente')
)

_ = trads.trad



from tinamit.mod import Modelo


class ModeloBF(Modelo):
    """
    La clase pariente para todos modelos biofísicos.
    """

    def __init__(símismo, variables, nombre='bf'):
        super().__init__(variables, nombre)

    @classmethod
    def prb_ingreso(cls):
        """
        Debe devolver la ubicación de un archivo de ingresos y una función que lo puede leer.

        Returns
        -------
        tuple[str, Callable]
        """

        pass

    @classmethod
    def prb_egreso(cls):
        """
        Debe devolver la ubicación de un archivo de egresos y una función que lo puede leer.

        Returns
        -------
        tuple[str, Callable]
        """

        pass

    @classmethod
    def prb_simul(cls):
        """
        Debe devolver la ubicación de un archivo de ingresos para correr una simulación de prueba.

        Returns
        -------
        str
        """
        pass

    def unidad_tiempo(símismo):
        raise NotImplementedError



from .._vars import VarConstante, VarAuxiliar, VarNivel, VarMDS, VariablesMDS


class VariablesPySD(VariablesMDS):
    pass


class VarPySD(VarMDS):
    def __init__(símismo, nombre_py, **args_ll):
        símismo.nombre_py = nombre_py
        super().__init__(**args_ll)


class VarPySDConstante(VarPySD, VarConstante):
    def __init__(símismo, nombre, nombre_py, unid, ec, parientes, inic=0, subs=None, líms=None, info=''):
        super().__init__(
            nombre_py, nombre=nombre, unid=unid, ec=ec, parientes=parientes, inic=inic, subs=subs, líms=líms,
            info=info
        )


class VarPySDAuxiliar(VarPySD, VarAuxiliar):

    def __init__(símismo, nombre, nombre_py, unid, ec, parientes, inic=0, subs=None, líms=None, info=''):
        super().__init__(
            nombre_py, nombre=nombre, unid=unid, ec=ec, parientes=parientes, inic=inic, subs=subs, líms=líms,
            info=info
        )


class VarPySDNivel(VarPySD, VarNivel):
    def __init__(símismo, nombre, nombre_py, unid, ec, parientes, inic=0, subs=None, líms=None, info=''):
        super().__init__(
            nombre_py, nombre=nombre, unid=unid, ec=ec, parientes=parientes, inic=inic, subs=subs, líms=líms,
            info=info
        )



import numpy as np

from tinamit.calibs._utils import calc_máx_trz
from tinamit.calibs.ec import CalibradorEc
from tinamit.config import _

try:
    import pymc3 as pm
except ImportError:  # pragma: sin cobertura
    pm = None


class CalibradorEcBayes(CalibradorEc):
    """
    Calibrador de ecuaciones con inferencia bayesiana.
    """

    def calibrar(
            símismo, bd, lugar=None, líms_paráms=None, ops=None, corresp_vars=None, ord_niveles=None, jerárquico=True
    ):
        """
        Efectua una calibración bayesiana para cada lugar en ``Lugar`` según los datos en ``bd``.

        Parameters
        ----------
        bd: BD
            La base de datos con observaciones para los variables en la ecuación.
        lugar: Lugar
            El lugar cuyos sublugares hay que calibrar; si es ``None`` se calibrará la ecuación con todos
            los datos en ``bd`` sin tener su lugar en cuenta.
        líms_paráms: list
            Límites teoréticos para los parámetros.
        ops: dict
            Opciones que se pasarán directamente a la función de calibración.
        corresp_vars: dict
            Diccionario de correspondencia entre los nombres de los variables en ``bd`` y sus nombres en la ecuación.
        ord_niveles: list
            Desambiguación del orden de niveles.
        jerárquico: bool
            Si empleamos inferencia bayesiana jerárquica o normal.

        Returns
        -------
        dict
            Diccionario con las calibraciones de cada lugar.
        """

        # Si no tenemos PyMC3, no podemos hacer inferencia bayesiana.
        if pm is None:
            raise ImportError(_('Debes instalar PyMC3 para poder hacer calibraciones con inferencia bayesiana.'))

        líms_paráms = símismo._gen_líms_paráms(líms_paráms)

        # Leer los variables de la ecuación.
        vars_x, var_y = símismo._extraer_vars()
        # Todas las observaciones
        obs = símismo._obt_datos(bd, vars_interés=vars_x + [var_y], corresp_vars=corresp_vars)
        obs = obs.dropna('n', how='any')

        if lugar is None:
            # Si no hay lugares, generar y calibrar el modelo de una vez.
            mod_bayes = símismo.ec.gen_mod_bayes(
                líms_paráms=líms_paráms, obs_x=obs[vars_x], obs_y=obs[var_y],
                binario=False, aprioris=None, nv_jerarquía=None
            )

            return _calibrar_mod_bayes(mod_bayes, paráms=list(símismo.paráms), ops=ops)

        # Si hay distribución geográfica, es un poco más complicado.
        if jerárquico:
            # Si estamos implementando un modelo jerárquico...

            ord_niveles = lugar.ord_niveles.resolver(ord_niveles)
            lugs = [lg for lg in lugar.lugares() if lg.nivel in ord_niveles]
            primer_nivel = ord_niveles[0]
            obs = obs.where(obs[_('lugar')].isin([
                h.cód for x in lugar.lugares(nivel=primer_nivel) for h in x
            ]), drop=True)

            # Primero, crear una lista de las relaciones jerárquicas, el cual se necesita para crear el modelo
            # jerárquico bayes.
            nvs_jerarq = [[lugar]]
            while len(lugs):
                for lug in nvs_jerarq[-1]:
                    lugs.remove(lug)
                lgs_nv = [lug for lug in lugs if lugar.pariente(lug) in nvs_jerarq[-1]]
                if len(lgs_nv):
                    nvs_jerarq.append(lgs_nv)
            for í, nv in enumerate(nvs_jerarq):
                nv[:] = [
                    lg for lg in nv
                    if (obs[_('lugar')].isin([x.cód for x in lg.lugares()]).sum()
                        and ((í == len(nvs_jerarq) - 1) or any(
                                lg is lugar.pariente(x, ord_niveles) for x in nvs_jerarq[í + 1]))
                        )
                ]
            í_nv_jerarquía = []
            for í, nv in enumerate(nvs_jerarq[1:]):
                í_nv_jerarquía.append([nvs_jerarq[í].index(lugar.pariente(lg, ord_niveles=ord_niveles)) for lg in nv])

            í_datos = np.array(
                [next(i for i, lg in enumerate(nvs_jerarq[-1]) if x == lg.cód or
                      any(p == lg for p in lugar.pariente(x, ord_niveles=ord_niveles, todos=True))) for x in
                 obs[_('lugar')].values.tolist()
                 ]
            )
            # Generar el modelo bayes
            mod_bayes_jrq = símismo.ec.gen_mod_bayes(
                líms_paráms=líms_paráms, obs_x=obs[vars_x], obs_y=obs[var_y],
                aprioris=None, binario=False, nv_jerarquía=í_nv_jerarquía,
                í_datos=í_datos
            )
            prms_extras = [
                'mu_{p}_nv_{í}'.format(í=í_nv, p=p) for í_nv in range(len(nvs_jerarq) - 1) for p in símismo.paráms
            ]

            # Calibrar
            res_calib = _calibrar_mod_bayes(
                mod_bayes_jrq, paráms=símismo.paráms + prms_extras, ops=ops
            )

            # Formatear los resultados
            resultados = {}
            for i, nv in enumerate(nvs_jerarq):
                for j, lg in enumerate(nv):
                    if i == (len(nvs_jerarq) - 1):
                        res_lg = {p: {ll: v[..., j] for ll, v in res_calib[p].items()} for p in símismo.paráms}
                    else:
                        res_lg = {
                            p: {ll: v[..., j] for ll, v in res_calib['mu_{}_nv_{}'.format(p, i)].items()}
                            for p in símismo.paráms
                        }
                    resultados[lg.cód] = res_lg

        else:
            # Si no estamos haciendo un modelo jerárquico, hay que calibrar cada lugar individualmente.

            # Efectuar las calibraciones para todos los lugares.
            resultados = {}
            sub_lugs = list(lugar.lugares())
            sub_lugs.remove(lugar)

            for lg in sub_lugs:
                obs_lg = obs.where(obs['lugar'].isin([x.cód for x in lg.lugares()]), drop=True)

                if len(obs_lg['n']):
                    mod_bayes = símismo.ec.gen_mod_bayes(
                        líms_paráms=líms_paráms, obs_x=obs_lg[vars_x], obs_y=obs_lg[var_y],
                        binario=False, aprioris=None, nv_jerarquía=None
                    )
                    resultados[lg.cód] = _calibrar_mod_bayes(
                        mod_bayes=mod_bayes, paráms=símismo.paráms, obs=obs_lg, ops=ops
                    )
        return resultados


# Unas funciones auxiliares
def _calibrar_mod_bayes(mod_bayes, paráms, obs=None, vars_compartidos=None, ops=None):
    """
    Esta función calibra un modelo bayes.

    Parameters
    ----------
    mod_bayes: pm.Modelo
        El modelo para calibrar.
    paráms: list[str]
        Una lista de los nombres de los parámetros de interés para sacar de la traza del modelo.
    obs: dict
        Base de datos observados.
    vars_compartidos: dict
        Un diccionario con los variables compartidos Theano en los cuales podemos poner nuevas observaciones.
    ops: dict
        Opciones adicionales para pasar a pm.Modelo.sample.

    Returns
    -------

    """

    # El diccionario de opciones adicionales.
    if ops is None:
        ops = {}

    # Si hay variables de datos compartidos, poner los nuevos datos.
    if vars_compartidos is not None:
        for var, var_pymc in vars_compartidos.items():
            var_pymc.set_value(obs[var])

    # Crear el diccionarion de argumentos
    ops_auto = {
        'tune': 1000,
        'nuts_kwargs': {'target_accept': 0.90}
    }
    ops_auto.update(ops)

    # Efectuar la calibración
    with mod_bayes:
        t = pm.sample(**ops_auto)

    # Devolver los datos procesados
    return _procesar_calib_bayes(t, paráms=paráms)


def _procesar_calib_bayes(traza, paráms):
    """
    Procesa los resultados de una calibración bayes. Con base en la traza PyMC3, calcula el punto de probabilidad más
    alta para cada parámetro de interés.

    Parameters
    ----------
    traza: pm.Trace
        La traza PyMC3.
    paráms: list
        La lista de parámetros de interés.

    Returns
    -------
    dict
        Los resultados procesados.
    """

    # El diccionario para los resultados
    d_máx = {}
    d_prom = {}

    # Calcular el punto de probabilidad máxima
    for p in paráms:
        # Para cada parámetro...

        dims = traza[p].shape
        if len(dims) == 1:
            d_máx[p] = calc_máx_trz(traza[p])
            d_prom[p] = np.mean(traza[p])
        elif len(dims) == 2:
            d_máx[p] = np.empty(dims[1])
            d_prom[p] = np.empty(dims[1])
            for e in range(dims[1]):
                d_máx[p][e] = calc_máx_trz(traza[p][:, e])
                d_prom[p][e] = np.mean(traza[p][:, e])
        else:
            raise ValueError

    # Devolver los resultados procesados.
    return {p: {'val': d_prom[p], 'cumbre': d_máx[p], 'dist': traza[p]} for p in paráms}



import csv
import os
from collections import OrderedDict

from tinamit.config import _
from tinamit.cositas import detectar_codif


class Nivel(object):
    """
    Un nivel geográfico (p. ej, ``municipio`` o ``departamento``.
    """
    def __init__(símismo, nombre, subniveles=None):
        """

        Parameters
        ----------
        nombre: str
            El nombre del nivel.
        subniveles: list of Nivel
            Lista de subniveles.
        """
        símismo.nombre = nombre
        símismo.subniveles = subniveles

    def __eq__(símismo, otro):
        return str(símismo) == str(otro)

    def __str__(símismo):
        return símismo.nombre

    def __getitem__(símismo, itema):
        return next(n for n in símismo.subniveles if n == itema)

    def __hash__(símismo):
        return hash(str(símismo))


class Lugar(object):
    """
    Un lugar dado en una geografía.
    """
    def __init__(símismo, nombre, nivel, cód=None, sub_lugares=None):
        """

        Parameters
        ----------
        nombre: str
            El nombre del lugar.
        nivel: Nivel
            El nivel geográfico correspondiente.
        cód: str
            El identificador único de este lugar. Si es ``None``, se tomará su nombre como identificador.
        sub_lugares:
            Lugares que se encuentre adentro de este.
        """
        símismo.cód = cód or nombre
        símismo.nivel = nivel
        símismo.nombre = nombre
        símismo.sub_lugares = set(sub_lugares or [])

        símismo.ord_niveles = _OrdNiveles(símismo)

    def lugares(símismo, en=None, nivel=None):
        """
        Devolver los sublugares presentes en este lugar.

        Parameters
        ----------
        en: str or Lugar
            Sublugar al cual limitir la búsqueda.
        nivel: Nivel or str or list
            Opción para limitir los resultados a uno o más niveles.

        Returns
        -------
        set[Lugar]
        """
        if isinstance(nivel, (str, Nivel)):
            nivel = [nivel]
        if en is None:
            buscar_en = símismo
        else:
            buscar_en = símismo[en]
        return {lg for lg in buscar_en if (nivel is None or lg.nivel in nivel)}

    def buscar_nombre(símismo, nombre, nivel=None):
        """
        Devuelve el sublugar con el nombre dado.

        Parameters
        ----------
        nombre: str
            El nombre del lugar deseado.
        nivel: Nivel or str
            Desambiguación en el caso que hayan múltiples lugares con el mismo nombre en distintos niveles.

        Returns
        -------
        Lugar

        """
        for lg in símismo:
            if lg.nombre == nombre and (nivel is None or lg.nivel == nivel):
                return lg
        raise ValueError(_('Lugar "{nmb}" no encontrado en "{lg}"').format(nmb=nombre, lg=símismo))

    def pariente(símismo, lugar, ord_niveles=None, todos=False):
        """
        Obtener el pariente de un sublugar dado.

        Parameters
        ----------
        lugar: str or Lugar
            Un sublugar cuyo pariente queremos.
        ord_niveles: list
            Desambiguación para lugares con niveles paralelos.
        todos: bool
            Si queremos todos los parientes del lugar, o solamente el más cercaco.

        Returns
        -------
        Lugar

        """
        lugar = símismo[lugar]
        ord_niveles = símismo.ord_niveles.resolver(ord_niveles)
        potenciales = [lg for lg in símismo if lugar in lg.sub_lugares and lg.nivel in ord_niveles]
        if potenciales:
            if todos:
                return sorted(potenciales, key=lambda x: ord_niveles.index(str(x.nivel)))
            return sorted(potenciales, key=lambda x: ord_niveles.index(str(x.nivel)))[0]

    def hijos_inmediatos(símismo, ord_niveles=None):
        """
        Devuelve los hijos inmediatos de este ``Lugar``.

        Parameters
        ----------
        ord_niveles: list
            Desambiguación para lugares con niveles paralelos.

        Returns
        -------
        list[Lugar]

        """

        return [lg for lg in símismo if símismo.pariente(lg, ord_niveles=ord_niveles) == símismo]

    def __iter__(símismo):
        yield símismo
        for lg in símismo.sub_lugares:
            for s_lg in lg:
                yield s_lg

    def __getitem__(símismo, itema):
        for lg in símismo:
            if isinstance(itema, Lugar) and lg is itema:
                return lg
            elif itema == lg.cód:
                return lg

        raise KeyError(itema)

    def __str__(símismo):
        return símismo.nombre


class _OrdNiveles(object):
    def __init__(símismo, lugar):

        sub_ords = list(set(sub.ord_niveles.ords for sub in lugar.sub_lugares))
        ords = []
        for sb in sub_ords:
            for i, nv in enumerate(sb):
                if i > (len(ords) - 1):
                    ords.append(nv)
                else:
                    ant = ords[i]
                    if isinstance(ant, Nivel) and ant != nv:
                        ords[i] = (ant, nv)
                    elif isinstance(ant, tuple) and nv not in ant:
                        ords[i] = (*ant, nv)

        símismo.ords = (*ords, lugar.nivel)

    def resolver(símismo, orden=None):
        if orden is None:
            return [x if isinstance(x, Nivel) else x[0] for x in símismo.ords]

        if isinstance(orden, (str, Nivel)):
            orden = [orden]
        return orden

    def __contains__(símismo, itema):
        for x in símismo.ords:
            if (isinstance(x, Nivel) and x == itema) or (itema in x):
                return True
        return False


def gen_lugares(archivo, nivel_base, nombre=None, col_cód='Código'):
    """
    Genera un lugar con todos los niveles y sublugares asociados desde un archivo ``.csv``.

    Cada columna en el ``.csv`` debe empezar con el nombre de un nivel, con la excepción de la columna ``col_cód``,
    la cual tendrá el código identificador único de cada lugar.

    Cada fila representa un lugar, con su **nombre** en la columna correspondiendo al nivel de este lugar y
    el **código** del lugar pariente en las otras columnas. Si un nivel no se aplica a un lugar (por ejemplo,
    un departamento no tendrá municipio pariente), se deja vacía la célula.

    Parameters
    ----------
    archivo: str
        El archivo ``.csv``.
    nivel_base: str
        El el nivel más alto. Por ejemplo, si tu csv entero representa un país, sería ``país``.
    nombre: str
        El nombre del lugar correspondiendo al nivel más alto. Por ejemplo, "Guatemala".
    col_cód: str
        El nombre de la columna con los códigos de cada sublugar.

    Returns
    -------
    Lugar
    """
    codif_csv = detectar_codif(archivo)
    nombre = nombre or os.path.splitext(os.path.split(nombre)[1])[0]

    with open(archivo, newline='', encoding=codif_csv) as d:

        lc = csv.DictReader(d)  # El lector de csv

        # Guardar la primera fila como nombres de columnas
        cols = [x.strip() for x in lc.fieldnames]

        if col_cód not in cols:
            raise ValueError(_(
                'La columna de código de región especificada ("{c}") no concuerda con los nombres de '
                'columnas del csv ({n}).'
            ).format(c=col_cód, n=', '.join(cols)))

        doc = [OrderedDict((ll.strip(), v.strip()) for ll, v in f.items()) for f in lc]

    # Inferir el orden de la jerarquía
    órden = []

    escalas = [x for x in cols if x != col_cód]

    coescalas = []
    for f in doc:
        coescalas_f = {ll for ll, v in f.items() if len(v) and ll in escalas}
        if not any(x == coescalas_f for x in coescalas):
            coescalas.append(coescalas_f)

    coescalas.sort(key=lambda x: len(x))

    while len(coescalas):
        siguientes = {x.pop() for x in coescalas if len(x) == 1}
        if not len(siguientes):
            raise ValueError(_('Parece que hay un error con el fuente de información regional.'))
        órden.append(sorted(list(siguientes)) if len(siguientes) > 1 else siguientes.pop())
        for cn in coescalas.copy():
            cn.difference_update(siguientes)
            if not len(cn):
                coescalas.remove(cn)

    órden.insert(0, nivel_base)
    niveles = {}
    anterior = None
    for nvls in órden[::-1]:
        if isinstance(nvls, str):
            niveles[nvls] = Nivel(nombre=nvls, subniveles=anterior)
            anterior = niveles[nvls]
        else:
            for nvl in nvls:
                niveles[nvl] = Nivel(nombre=nvl, subniveles=anterior)
            anterior = [niveles[nv] for nv in nvls]

    dic_doc = {f[col_cód]: f for f in doc}
    quedan = list(dic_doc)
    lugares = {}
    for nvl, obj_nvl in niveles.items():
        for cód in list(quedan):
            lg = dic_doc[cód]
            if lg[nvl]:
                nmb = lg[nvl]
                subs = {s for c, s in lugares.items() if dic_doc[c][nvl] == cód}
                lugares[cód] = Lugar(nmb, nivel=obj_nvl, cód=cód, sub_lugares=subs)
                quedan.remove(cód)

    return Lugar(nombre=nombre, nivel=niveles[nivel_base], sub_lugares=lugares.values())
